/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.input;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.*;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.filter.ElementFilter;
import org.jdom.input.SAXBuilder;
import monalisa.MonaLisa;
import monalisa.data.Pair;
import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.gui.KeggDialog;
import monalisa.resources.ResourceManager;
import monalisa.util.FileUtils;

/**
 * Import KEGG networks in xml-format.
 * http://www.genome.jp/kegg/docs/xml/
 * There are interaction (regulation) and reaction (metabolism) networks.
 * 
 * entry: - id
 *        - name
 *        - type
 *        - reaction
 *        - graphics: - name
 *                    - x, y, width, height
 * relation: - entry1
 *           - entry2
 *           - type
 *           - subtype: - name
 *                      - value
 * reaction: - name
 *           - type
 *           - substrate: - name
 *           - product: - name
 * 
 * @author Anja Thormann
 */
public class KeggInputHandler implements InputHandler {
    
    private final PetriNet pn = new PetriNet();
    private final Map<Integer, Transition> transitions = new HashMap<Integer, Transition>();
    private final Map<String, Place> places = new HashMap<String, Place>();
    private final Map<String, Element> key2Entry = new HashMap<String, Element>();
    private final Map<String, Pair<Double, Double> > id2pos = new HashMap<String, Pair<Double, Double> >();
    private final List<Element> groups = new ArrayList<Element>();
    private int countPlaces = -1;
    private int countTransitions = -1;
    private HashMap<String, String> keggCompoundDefinitions;
    
    @Override
    public boolean isKnownFile(File file) throws IOException {
        if(!"xml".equalsIgnoreCase(FileUtils.getExtension(file)))
            return false;
        try {
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(file);
            Element root = doc.getRootElement();
            String image = root.getAttributeValue("image");
            return image != null && image.contains("kegg");
        } catch(JDOMException e){
            return false;
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public PetriNet load(InputStream in) throws IOException {
        try {
            keggCompoundDefinitions = readDefinitions(ResourceManager.instance().getResourceUrl("kegg/keggCompoundDefinition.txt"));
        }
        catch (IOException e) {
            throw new IOException("Unable to read KEGG compound definition file", e);
        }

        SAXBuilder builder = new SAXBuilder();
        Document doc;
        try {
            doc = builder.build(in);
        } catch (JDOMException e) {
            throw new IOException("Failed to parse the XML file.", e);
        }
        Element root = doc.getRootElement();

        List<Element> entries = findElements(root, "entry");
        List<Element> relations = findElements(root, "relation");
        List<Element> reactions = findElements(root, "reaction");
        
        String id, name, type;
        Double x, y;

        for(Element entry : entries){
            id = entry.getAttributeValue("id");
            name = entry.getAttributeValue("name");
            type = entry.getAttributeValue("type"); // gene, group, compound, (map, enzyme) x, y, width, height
            x = 50.0;
            y = 50.0;
            // "x" and "y" are not required attributes in KGML
            try {
                x = new Double(entry.getChild("graphics").getAttributeValue("x"));
                y = new Double(entry.getChild("graphics").getAttributeValue("y"));
            } catch(Exception e) {

            }
            if(type.equals("group")){
                groups.add(entry);
            }
            else if(type.equals("compound")){
                String compoundName = keggCompoundDefinitions.get(entry.getChild("graphics").getAttributeValue("name"));
                if(compoundName != null){
                    entry.getChild("graphics").setAttribute("name", compoundName);
                }
                key2Entry.put(name, entry);
                key2Entry.put(id, entry);
                id2pos.put(name, new Pair<Double, Double>(x,y));
                id2pos.put(id, new Pair<Double, Double>(x,y));
            }
            else{
                key2Entry.put(id, entry);
            }           
        }

        for(Element group : groups){
            List<Element> components = group.getChildren("component");
            name = "complex";
            for(Element c : components){
                Element entry = key2Entry.get(c.getAttributeValue("id"));
                name = name.concat("_"+entry.getChild("graphics").getAttributeValue("name").replace("..", ""));
            }
            group.setAttribute("name", name);
            key2Entry.put(group.getAttributeValue("id"), group);
            key2Entry.put(group.getAttributeValue("name"), group);
        }

        // ECrel, PPrel, GErel, PCrel, maplink
        for(Element relation : relations){
            type = relation.getAttributeValue("type");
            if(!type.equals("ECrel") && !type.equals("maplink")){
                String entry1 = relation.getAttributeValue("entry1");
                String entry2 = relation.getAttributeValue("entry2");
                String subtype = relation.getChild("subtype").getAttributeValue("name");

                Place place1 = findPlace(entry1);
                Place place2 = findPlace(entry2);

                countTransitions++;
                Transition transition = findTransition(countTransitions, pn);
                transition.putProperty("name", subtype);
                pn.addArc(place1, transition);
                pn.addArc(transition, place2);
            }
        }

        for (Element reaction : reactions) {
            name = reaction.getAttributeValue("name");
            type = reaction.getAttributeValue("type");
            List<Element> substrates = findElements(reaction, "substrate");
            List<Element> products = findElements(reaction, "product");

            if (type.equals("reversible")) {
                countTransitions++;
                Transition forwardTransition = findTransition(countTransitions, pn);
                forwardTransition.putProperty("name", "F_" + name.replace("rn:", ""));

                countTransitions++;
                Transition backwardTransition = findTransition(countTransitions, pn);
                backwardTransition.putProperty("name", "B_" + name.replace("rn:", ""));


                for (Element s : substrates) {
                    String substrateName = s.getAttributeValue("name");
                    Place place = findPlace(substrateName);
                    pn.addArc(place, forwardTransition);
                    pn.addArc(backwardTransition, place);
                }
                for (Element p : products) {
                    String productName = p.getAttributeValue("name");
                    Place place = findPlace(productName);
                    pn.addArc(forwardTransition, place);
                    pn.addArc(place, backwardTransition);
                }
            } else {
                countTransitions++;
                Transition transition = findTransition(countTransitions, pn);
                transition.putProperty("name", name.replace("rn:", ""));

                for (Element s : substrates) {
                    String substrateName = s.getAttributeValue("name");
                    Place place = findPlace(substrateName);
                    pn.addArc(place, transition);
                }
                for (Element p : products) {
                    String productName = p.getAttributeValue("name");
                    Place place = findPlace(productName);
                    pn.addArc(transition, place);
                }
            }
        }
        
//        KeggDialog dialog = new KeggDialog(MonaLisa.appMainWindow(), pn.places());
//
//        for(Place p : dialog.inputs()){
//            countTransitions++;
//            Transition transition = findTransition(countTransitions, pn);
//            transition.putProperty("name", "input");
//            pn.addArc(transition, p);
//        }
//
//        for(Place p : dialog.outputs()){
//            countTransitions++;
//            Transition transition = findTransition(countTransitions, pn);
//            transition.putProperty("name", "output");
//            pn.addArc(p, transition);
//        }

        return pn;
    }
    
   
    private Transition findTransition(int transitionId, PetriNet petriNet) {
        Transition transition = transitions.get(transitionId);
        
        if (transition == null) {
            transition = new Transition(transitionId);
            transitions.put(transitionId, transition);
            petriNet.addTransition(transition);
        }
        return transition;
    }
    
    private Place findPlace(String key) {
        Element entry = key2Entry.get(key);
        Place place = places.get(key);
//        Pair<Double, Double> pos;

        if(place == null){
            place = new Place(++countPlaces);
            String name = entry.getChild("graphics").getAttributeValue("name");
            if(name == null){
                name = entry.getAttributeValue("name");
            }
            place.putProperty("name", name.replace("...", ""));

//            pos = id2pos.get(key);
//            place.putProperty("posX", pos.first());
//            place.putProperty("posY", pos.second());

            pn.addPlace(place);
            places.put(key, place);
        }
        return place;
    }
    
    @SuppressWarnings("unchecked")
    private List<Element> findElements(Element node, String request) {
        List<Element> returnList = new ArrayList<Element>();
        Iterator<Element> iterator = node.getDescendants(new ElementFilter());
        while (iterator.hasNext()) {
            Element elem = iterator.next();
            if (elem.getName().equals(request)) {
                returnList.add(elem);
            }
        }
        return returnList;
    }
    
    public HashMap<String, String> readDefinitions(URL resource) throws IOException {
        HashMap<String, String> ret = new HashMap<String, String>();
        InputStream istream = resource.openStream();
        DataInputStream in = null;
        try {
            in = new DataInputStream(istream);
            BufferedReader br = new BufferedReader(new InputStreamReader(in));
            String line;
            StringTokenizer st;
            while ((line = br.readLine()) != null){
                st = new StringTokenizer(line);
                String name = st.nextToken();
                String definition = "";
                while(st.hasMoreTokens()){
                    definition = definition.concat(st.nextToken()+" ");
                }
                String[] split = definition.split(";");
                ret.put(name, split[0].trim());
            }
        }
        finally {
            if (in != null)
                in.close();
        }
        return ret;
    }
}