/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.output;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.stream.XMLStreamException;

import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.resources.ResourceManager;
import monalisa.resources.StringResources;
import monalisa.util.FileUtils;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

import org.sbml.jsbml.Compartment;
import org.sbml.jsbml.Model;
import org.sbml.jsbml.Reaction;
import org.sbml.jsbml.SBMLDocument;
import org.sbml.jsbml.SBMLException;
import org.sbml.jsbml.SBMLWriter;
import org.sbml.jsbml.Species;
import org.sbml.jsbml.SpeciesReference;

/**
 *
 * @author Jens Einloft
 */
public class SbmlOutputHandler implements OutputHandler {

    private static final ResourceManager resources = ResourceManager.instance();
    private static final StringResources strings = resources.getDefaultStrings();

    public void save(FileOutputStream fileOutputStream, PetriNet pn) {
        SBMLDocument doc = new SBMLDocument(3,1);
        Model model = doc.createModel("MonaLisaExport");

        Compartment compartment = model.createCompartment("default");
        compartment.setSize(new Double(1.0));
        compartment.setConstant(true);

        Species species = null;
        for(Place p : pn.places()) {
            species = model.createSpecies("P"+p.id(), compartment);
            species.setName((String) p.getProperty("name"));
            species.setCompartment(compartment);
            species.setHasOnlySubstanceUnits(true);
            species.setBoundaryCondition(false);
            species.setConstant(false);

            if(p.hasProperty("token"))
                species.setInitialAmount((Double) p.getProperty("token"));
            else
                species.setInitialAmount(new Double(0.0));                     
        }

        Reaction reaction = null;
        for(Transition t : pn.transitions()) {
            reaction = model.createReaction();

            reaction.setId("T"+t.id());
            reaction.setName((String) t.getProperty("name"));
            reaction.setReversible(false);
            reaction.setFast(false);

            SpeciesReference sr = null;
            for(Place p : t.inputs()) {
                sr = reaction.createReactant(model.getSpecies("P"+p.id()));
                sr.setStoichiometry(pn.getArc(p, t).weight());
                sr.setConstant(true);
            }

            for(Place p : t.outputs()) {
                sr = reaction.createProduct(model.getSpecies("P"+p.id()));
                sr.setStoichiometry(pn.getArc(t, p).weight());
                sr.setConstant(true);
            }                
        }

        SBMLWriter writer = new SBMLWriter();
        try {
            writer.write(doc, fileOutputStream, "MonaLisa", strings.get("CurrentVersion"));
        } catch (XMLStreamException ex) {
            Logger.getLogger(SbmlOutputHandler.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SBMLException ex) {
            Logger.getLogger(SbmlOutputHandler.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public boolean isKnownFile(File file) throws IOException {
        if ("sbml".equalsIgnoreCase(FileUtils.getExtension(file)))
            return true;

        if("xml".equalsIgnoreCase(FileUtils.getExtension(file))) {
            SAXBuilder builder = new SAXBuilder();
            Document doc = null;
            try {
                doc = builder.build(file);
            } catch (Exception e) {
                return false;
            }

            Element root = doc.getRootElement();
            if (!root.getName().equals("sbml"))
                return false;
            else
                return true;
        }
        return false;
    }

    public File checkFileNameForExtension(File file) {
        if(!"xml".equalsIgnoreCase(FileUtils.getExtension(file)))
            file = new File(file.getAbsolutePath()+".xml");
        return file;
    }

    public String getExtension() {
        return "xml";
    }

    public String getDescription() {
        return "SBML";
    }
}