/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.distance;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.*;
import javax.swing.border.TitledBorder;

import monalisa.data.Pair;
import monalisa.data.pn.TInvariant;
import monalisa.resources.StringResources;
import monalisa.results.Configuration;
import monalisa.results.DistanceMatrix;
import monalisa.results.DistanceMatrixConfiguration;
import monalisa.results.TInvariants;
import monalisa.results.TInvariantsConfiguration;
import monalisa.tools.AbstractTool;
import monalisa.tools.ErrorLog;
import monalisa.tools.Tool;
import monalisa.tools.tinv.TInvariantTool;
import monalisa.util.Components;
import monalisa.Project;

public final class DistanceTool extends AbstractTool implements ActionListener {
    private static final String ACTION_CALCULATE = "CALCULATE";
    private JPanel panel;
    private JPanel distanceFctPanel;
    private JCheckBox tanimotoCheckbox;
    private JCheckBox sumofdiffCheckbox;
    private JCheckBox simplematchingCheckbox;
    private JCheckBox includeTrivialTinvCheckbox;
    private JCheckBox calculateButton;
    private int totalNumberOfCalculations;
    private int currentCalculation;
    private Project project;

    @Override
    public void run(Project project, ErrorLog log) throws InterruptedException {
        boolean includeTrivialTInv = includeTrivialTinvCheckbox.isSelected();
        
        totalNumberOfCalculations = getActiveDistanceFunctions().size();
        currentCalculation = 0;
        
        for (String function : getActiveDistanceFunctions()) {
            computeDistanceMatrix(project, function, includeTrivialTInv);
            currentCalculation++;
        }

        enableExport(true);
    }

    private List<String> getActiveDistanceFunctions() {
        List<String> result = new ArrayList<String>();
        
        if (sumofdiffCheckbox.isSelected())
            result.add(DistanceFunctions.SUM_OF_DIFFERENCES);
        if (tanimotoCheckbox.isSelected())
            result.add(DistanceFunctions.TANIMOTO);
        if (simplematchingCheckbox.isSelected())
            result.add(DistanceFunctions.SIMPLEMATCHING);
        
        return result;
    }
    
    @Override
    public boolean finishedState(Project project) {
        if (project.hasAllResults(this, 6)) {
            calculateButton.setSelected(false);
            Components.setEnabled(panel, false);
            return true;
        }
        return false;
    }

    @Override
    public JPanel getUI(Project project, StringResources strings) {
        this.project = project;
        if (panel == null) {
            exportButton.setActionCommand(EXPORT);
            exportButton.addActionListener(this);
            enableExport(project.hasResults(this));

            tanimotoCheckbox = new JCheckBox(DistanceFunctions.getName(DistanceFunctions.TANIMOTO));
            tanimotoCheckbox.setSelected(true);
            sumofdiffCheckbox = new JCheckBox(DistanceFunctions.getName(DistanceFunctions.SUM_OF_DIFFERENCES));
            simplematchingCheckbox = new JCheckBox(DistanceFunctions.getName(DistanceFunctions.SIMPLEMATCHING));
            
            includeTrivialTinvCheckbox = new JCheckBox(strings.get("IncludeTrivialTInvariants"));
            includeTrivialTinvCheckbox.setSelected(false);
            
            calculateButton = new JCheckBox(strings.get("Calculate"));
            calculateButton.setActionCommand(ACTION_CALCULATE);
            calculateButton.addActionListener(this);
            
            distanceFctPanel = new JPanel();
            distanceFctPanel.setBorder(BorderFactory.createTitledBorder(
                null, strings.get("DistanceFunctions"),
                TitledBorder.DEFAULT_JUSTIFICATION,
                TitledBorder.DEFAULT_POSITION));
            
            GroupLayout distanceFctLayout = new GroupLayout(distanceFctPanel);
            distanceFctPanel.setLayout(distanceFctLayout);
            
            distanceFctLayout.setHorizontalGroup(distanceFctLayout.createParallelGroup()
                .addComponent(tanimotoCheckbox)
                .addComponent(sumofdiffCheckbox)
                .addComponent(simplematchingCheckbox));
            distanceFctLayout.setVerticalGroup(distanceFctLayout.createSequentialGroup()
                .addComponent(tanimotoCheckbox)
                .addComponent(sumofdiffCheckbox)
                .addComponent(simplematchingCheckbox));
            
            panel = new JPanel();
            GroupLayout layout = new GroupLayout(panel);
            panel.setLayout(layout);
            
            layout.setAutoCreateGaps(true);
            layout.setAutoCreateContainerGaps(true);
            
            layout.setHorizontalGroup(layout.createParallelGroup()
                .addComponent(distanceFctPanel)
                .addComponent(includeTrivialTinvCheckbox)
                .addComponent(calculateButton));
//                .addComponent(exportButton));
            
            layout.setVerticalGroup(layout.createSequentialGroup()
                .addComponent(distanceFctPanel)
                .addComponent(includeTrivialTinvCheckbox)
                .addComponent(calculateButton));
//                .addComponent(exportButton));
        }

        return panel;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        final String action = e.getActionCommand();

        if (action == ACTION_CALCULATE)
            fireActivityChanged(isActive());
        else if(action == EXPORT) {
            List<Class<? extends Tool>> toolList = new ArrayList<Class<? extends Tool>>();
            toolList.add(DistanceTool.class);
            try {
                this.project.exportResults(toolList);
            } catch (IOException ex) {
                Logger.getLogger(DistanceTool.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    @Override
    public boolean isActive() {
        return calculateButton.isSelected();
    }

    @Override
    public void saveSettings(Project p) {
    // TODO Auto-generated method stub
    }

    @Override
    public void setActive(boolean active) {
        if (active != isActive()) {
            calculateButton.setSelected(active);
            fireActivityChanged(active);
        }
    }

    @Override
    public void setActive(Configuration... configs) {
        for (Configuration config : configs) {
            DistanceMatrixConfiguration dmconfig =
                (DistanceMatrixConfiguration) config;
            String dfname = dmconfig.getDistanceFunctionName();
            
            if (dfname == DistanceFunctions.SUM_OF_DIFFERENCES)
                sumofdiffCheckbox.setSelected(true);
            else if (dfname == DistanceFunctions.TANIMOTO)
                tanimotoCheckbox.setSelected(true);
            else if (dfname == DistanceFunctions.SIMPLEMATCHING)
                simplematchingCheckbox.setSelected(true);
        }
        
        if (!isActive())
            setActive(true);
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<Pair<Class<? extends Tool>, Configuration>> getRequirements() {
        if (isActive())
            return Arrays.asList(new Pair<Class<? extends Tool>, Configuration>(
                TInvariantTool.class, new TInvariantsConfiguration()));
        else
            return Collections.emptyList();
    }

    private void computeDistanceMatrix(Project project, String distFunction,
            boolean includeTrivialTInv) throws InterruptedException {
        DistanceFunctions functions = new DistanceFunctions();
        DistanceFunction df = functions.getDistanceFunc(distFunction);

        TInvariants allTinvariants = project.getResult(
            TInvariantTool.class, new TInvariantsConfiguration());
        List<TInvariant> tinvariants = new ArrayList<TInvariant>(
            includeTrivialTInv ? allTinvariants :
                allTinvariants.nonTrivialTInvariants());

        float[][] matrix = calculateDistanceMatrix(tinvariants, df);
        DistanceMatrixConfiguration config = new DistanceMatrixConfiguration(distFunction, includeTrivialTInv);
        addResult(config, new DistanceMatrix(matrix, distFunction, tinvariants));
    }

    /**
     * Taken from {@link Calculation#calculateDistanceMatrix(List, DistanceFunction)}.
     * @param tInvariants
     * @param distanceFunc
     * @return
     * @throws InterruptedException 
     */
    private float[][] calculateDistanceMatrix(
            List<TInvariant> tinvariants, DistanceFunction distanceFunc) throws
            InterruptedException {
        int n = tinvariants.size();
        float[][] ret = new float[n][n];

        for (int i = 0; i < n; i++) {
            if (Thread.interrupted())
                throw new InterruptedException();
            
            for (int j = i; j < n; j++) {
                float dist = distanceFunc.getDistanceValue(tinvariants.get(i),
                    tinvariants.get(j));
                ret[i][j] = dist;
                ret[j][i] = dist;
            }
            int progress = (i + n * currentCalculation) * 100 / (n * totalNumberOfCalculations);
            fireProgressUpdated(progress);
        }
        return ret;
    }
}
